use crate::parser::{Parser, ParserContext, ParserResult};
use std::cell::RefCell;
use std::fmt::Debug;
use std::rc::Rc;

pub trait ParserTuple<'a, TToken, T: 'a, TContext>
where
    TToken: Debug + Clone,
{
    fn parse(
        &self,
        context: Rc<RefCell<ParserContext<TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T>;
}

pub struct TupleParser<P> {
    pub(crate) parser: P,
}

impl<'a, TToken, T: 'a, TContext, P> Parser<'a, TToken, T, TContext> for TupleParser<P>
where
    TToken: Debug + Clone,
    P: ParserTuple<'a, TToken, T, TContext>,
{
    fn parse(
        &self,
        context: Rc<RefCell<ParserContext<TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T> {
        self.parser.parse(context, input)
    }
}

impl<'a, TToken, T1: 'a, T2: 'a, TContext, P1, P2> ParserTuple<'a, TToken, (T1, T2), TContext>
    for (P1, P2)
where
    TToken: Debug + Clone,
    P1: Parser<'a, TToken, T1, TContext>,
    P2: Parser<'a, TToken, T2, TContext>,
{
    fn parse(
        &self,
        context: Rc<RefCell<ParserContext<TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, (T1, T2)> {
        let (input, r1) = self.0.parse(context.clone(), input)?;
        let (input, r2) = self.1.parse(context, input)?;

        Ok((input, (r1, r2)))
    }
}

impl<'a, TToken, T1: 'a, T2: 'a, T3: 'a, TContext, P1, P2, P3>
    ParserTuple<'a, TToken, (T1, T2, T3), TContext> for (P1, P2, P3)
where
    TToken: Debug + Clone,
    P1: Parser<'a, TToken, T1, TContext>,
    P2: Parser<'a, TToken, T2, TContext>,
    P3: Parser<'a, TToken, T3, TContext>,
{
    fn parse(
        &self,
        context: Rc<RefCell<ParserContext<TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, (T1, T2, T3)> {
        let (input, r1) = self.0.parse(context.clone(), input)?;
        let (input, r2) = self.1.parse(context.clone(), input)?;
        let (input, r3) = self.2.parse(context.clone(), input)?;

        Ok((input, (r1, r2, r3)))
    }
}

impl<'a, TToken, T1: 'a, T2: 'a, T3: 'a, T4: 'a, TContext, P1, P2, P3, P4>
    ParserTuple<'a, TToken, (T1, T2, T3, T4), TContext> for (P1, P2, P3, P4)
where
    TToken: Debug + Clone,
    P1: Parser<'a, TToken, T1, TContext>,
    P2: Parser<'a, TToken, T2, TContext>,
    P3: Parser<'a, TToken, T3, TContext>,
    P4: Parser<'a, TToken, T4, TContext>,
{
    fn parse(
        &self,
        context: Rc<RefCell<ParserContext<TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, (T1, T2, T3, T4)> {
        let (input, r1) = self.0.parse(context.clone(), input)?;
        let (input, r2) = self.1.parse(context.clone(), input)?;
        let (input, r3) = self.2.parse(context.clone(), input)?;
        let (input, r4) = self.3.parse(context.clone(), input)?;

        Ok((input, (r1, r2, r3, r4)))
    }
}

impl<'a, TToken, T1: 'a, T2: 'a, T3: 'a, T4: 'a, T5: 'a, TContext, P1, P2, P3, P4, P5>
    ParserTuple<'a, TToken, (T1, T2, T3, T4, T5), TContext> for (P1, P2, P3, P4, P5)
where
    TToken: Debug + Clone,
    P1: Parser<'a, TToken, T1, TContext>,
    P2: Parser<'a, TToken, T2, TContext>,
    P3: Parser<'a, TToken, T3, TContext>,
    P4: Parser<'a, TToken, T4, TContext>,
    P5: Parser<'a, TToken, T5, TContext>,
{
    fn parse(
        &self,
        context: Rc<RefCell<ParserContext<TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, (T1, T2, T3, T4, T5)> {
        let (input, r1) = self.0.parse(context.clone(), input)?;
        let (input, r2) = self.1.parse(context.clone(), input)?;
        let (input, r3) = self.2.parse(context.clone(), input)?;
        let (input, r4) = self.3.parse(context.clone(), input)?;
        let (input, r5) = self.4.parse(context.clone(), input)?;

        Ok((input, (r1, r2, r3, r4, r5)))
    }
}

impl<
        'a,
        TToken,
        T1: 'a,
        T2: 'a,
        T3: 'a,
        T4: 'a,
        T5: 'a,
        T6: 'a,
        TContext,
        P1,
        P2,
        P3,
        P4,
        P5,
        P6,
    > ParserTuple<'a, TToken, (T1, T2, T3, T4, T5, T6), TContext> for (P1, P2, P3, P4, P5, P6)
where
    TToken: Debug + Clone,
    P1: Parser<'a, TToken, T1, TContext>,
    P2: Parser<'a, TToken, T2, TContext>,
    P3: Parser<'a, TToken, T3, TContext>,
    P4: Parser<'a, TToken, T4, TContext>,
    P5: Parser<'a, TToken, T5, TContext>,
    P6: Parser<'a, TToken, T6, TContext>,
{
    fn parse(
        &self,
        context: Rc<RefCell<ParserContext<TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, (T1, T2, T3, T4, T5, T6)> {
        let (input, r1) = self.0.parse(context.clone(), input)?;
        let (input, r2) = self.1.parse(context.clone(), input)?;
        let (input, r3) = self.2.parse(context.clone(), input)?;
        let (input, r4) = self.3.parse(context.clone(), input)?;
        let (input, r5) = self.4.parse(context.clone(), input)?;
        let (input, r6) = self.5.parse(context.clone(), input)?;

        Ok((input, (r1, r2, r3, r4, r5, r6)))
    }
}

impl<
        'a,
        TToken,
        T1: 'a,
        T2: 'a,
        T3: 'a,
        T4: 'a,
        T5: 'a,
        T6: 'a,
        T7: 'a,
        TContext,
        P1,
        P2,
        P3,
        P4,
        P5,
        P6,
        P7,
    > ParserTuple<'a, TToken, (T1, T2, T3, T4, T5, T6, T7), TContext>
    for (P1, P2, P3, P4, P5, P6, P7)
where
    TToken: Debug + Clone,
    P1: Parser<'a, TToken, T1, TContext>,
    P2: Parser<'a, TToken, T2, TContext>,
    P3: Parser<'a, TToken, T3, TContext>,
    P4: Parser<'a, TToken, T4, TContext>,
    P5: Parser<'a, TToken, T5, TContext>,
    P6: Parser<'a, TToken, T6, TContext>,
    P7: Parser<'a, TToken, T7, TContext>,
{
    fn parse(
        &self,
        context: Rc<RefCell<ParserContext<TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, (T1, T2, T3, T4, T5, T6, T7)> {
        let (input, r1) = self.0.parse(context.clone(), input)?;
        let (input, r2) = self.1.parse(context.clone(), input)?;
        let (input, r3) = self.2.parse(context.clone(), input)?;
        let (input, r4) = self.3.parse(context.clone(), input)?;
        let (input, r5) = self.4.parse(context.clone(), input)?;
        let (input, r6) = self.5.parse(context.clone(), input)?;
        let (input, r7) = self.6.parse(context.clone(), input)?;

        Ok((input, (r1, r2, r3, r4, r5, r6, r7)))
    }
}

impl<
        'a,
        TToken,
        T1: 'a,
        T2: 'a,
        T3: 'a,
        T4: 'a,
        T5: 'a,
        T6: 'a,
        T7: 'a,
        T8: 'a,
        TContext,
        P1,
        P2,
        P3,
        P4,
        P5,
        P6,
        P7,
        P8,
    > ParserTuple<'a, TToken, (T1, T2, T3, T4, T5, T6, T7, T8), TContext>
    for (P1, P2, P3, P4, P5, P6, P7, P8)
where
    TToken: Debug + Clone,
    P1: Parser<'a, TToken, T1, TContext>,
    P2: Parser<'a, TToken, T2, TContext>,
    P3: Parser<'a, TToken, T3, TContext>,
    P4: Parser<'a, TToken, T4, TContext>,
    P5: Parser<'a, TToken, T5, TContext>,
    P6: Parser<'a, TToken, T6, TContext>,
    P7: Parser<'a, TToken, T7, TContext>,
    P8: Parser<'a, TToken, T8, TContext>,
{
    fn parse(
        &self,
        context: Rc<RefCell<ParserContext<TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, (T1, T2, T3, T4, T5, T6, T7, T8)> {
        let (input, r1) = self.0.parse(context.clone(), input)?;
        let (input, r2) = self.1.parse(context.clone(), input)?;
        let (input, r3) = self.2.parse(context.clone(), input)?;
        let (input, r4) = self.3.parse(context.clone(), input)?;
        let (input, r5) = self.4.parse(context.clone(), input)?;
        let (input, r6) = self.5.parse(context.clone(), input)?;
        let (input, r7) = self.6.parse(context.clone(), input)?;
        let (input, r8) = self.7.parse(context.clone(), input)?;

        Ok((input, (r1, r2, r3, r4, r5, r6, r7, r8)))
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::combinators::tuple;
    use crate::parser::parser_test_helper;
    use crate::text::char_parser;

    #[test]
    fn tuple_test() {
        parser_test_helper("ab", &('a', 'b'), |c, i| {
            tuple((char_parser('a'), char_parser('b'))).parse(c, i)
        });

        parser_test_helper("abc", &('a', 'b', 'c'), |c, i| {
            tuple((char_parser('a'), char_parser('b'), char_parser('c'))).parse(c, i)
        });

        parser_test_helper("abcd", &('a', 'b', 'c', 'd'), |c, i| {
            tuple((
                char_parser('a'),
                char_parser('b'),
                char_parser('c'),
                char_parser('d'),
            ))
            .parse(c, i)
        });
    }
}
